@@ -90,6 +90,8 @@ struct wl_display {
 	struct wl_list global_list;
 	struct wl_list socket_list;
 	struct wl_list client_list;
+
+	struct wl_signal destroy_signal;
 };
 
 struct wl_global {
@@ -1078,7 +1080,7 @@ wl_display_create(void)
 	const char *debug;
 
 	debug = getenv("WAYLAND_DEBUG");
-	if (debug)
+	if (debug && (strstr(debug, "server") || strstr(debug, "1")))
 		wl_debug = 1;
 
 	display = malloc(sizeof *display);
@@ -1096,6 +1098,8 @@ wl_display_create(void)
 	wl_list_init(&display->client_list);
 	wl_list_init(&display->registry_resource_list);
 
+	wl_signal_init(&display->destroy_signal);
+
 	display->id = 1;
 	display->serial = 0;
 
@@ -1115,6 +1119,8 @@ wl_display_destroy(struct wl_display *display)
 	struct wl_socket *s, *next;
 	struct wl_global *global, *gnext;
 
+	wl_signal_emit(&display->destroy_signal, display);
+
 	wl_list_for_each_safe(s, next, &display->socket_list, link) {
 		wl_event_source_remove(s->source);
 		unlink(s->addr.sun_path);
@@ -1381,6 +1387,20 @@ wl_display_add_socket(struct wl_display *display, const char *name)
 	return 0;
 }
 
+WL_EXPORT void
+wl_display_add_destroy_listener(struct wl_display *display,
+				struct wl_listener *listener)
+{
+	wl_signal_add(&display->destroy_signal, listener);
+}
+
+WL_EXPORT struct wl_listener *
+wl_display_get_destroy_listener(struct wl_display *display,
+				wl_notify_func_t notify)
+{
+	return wl_signal_get(&display->destroy_signal, notify);
+}
+
 WL_EXPORT struct wl_resource *
 wl_client_add_object(struct wl_client *client,
 		     const struct wl_interface *interface,
