@@ -25,102 +25,72 @@
  *
  */
 
+#define _GNU_SOURCE
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/mman.h>
 #include <unistd.h>
 
+#include "wayland-private.h"
 #include "wayland-server.h"
 
-struct wl_shm {
-	const struct wl_shm_callbacks *callbacks;
+struct wl_shm_pool {
+	struct wl_resource resource;
+	int refcount;
+	char *data;
+	int size;
 };
 
 struct wl_shm_buffer {
 	struct wl_buffer buffer;
-	struct wl_shm *shm;
 	int32_t stride;
 	uint32_t format;
-	void *data;
+	int offset;
+	struct wl_shm_pool *pool;
 };
 
+static void
+shm_pool_unref(struct wl_shm_pool *pool)
+{
+	pool->refcount--;
+	if (pool->refcount)
+		return;
+
+	munmap(pool->data, pool->size);
+	free(pool);
+}
+
 static void
 destroy_buffer(struct wl_resource *resource)
 {
 	struct wl_shm_buffer *buffer =
 		container_of(resource, struct wl_shm_buffer, buffer.resource);
 
-	munmap(buffer->data, buffer->stride * buffer->buffer.height);
-
-	buffer->shm->callbacks->buffer_destroyed(&buffer->buffer);
-
+	if (buffer->pool)
+		shm_pool_unref(buffer->pool);
 	free(buffer);
 }
 
-static void
-shm_buffer_damage(struct wl_client *client, struct wl_resource *resource,
-		  int32_t x, int32_t y, int32_t width, int32_t height)
-{
-	struct wl_shm_buffer *buffer = resource->data;
-
-	buffer->shm->callbacks->buffer_damaged(&buffer->buffer, x, y,
-					       width, height);
-}
-
 static void
 shm_buffer_destroy(struct wl_client *client, struct wl_resource *resource)
 {
-	wl_resource_destroy(resource, 0);
+	wl_resource_destroy(resource);
 }
 
-const static struct wl_buffer_interface shm_buffer_interface = {
-	shm_buffer_damage,
+static const struct wl_buffer_interface shm_buffer_interface = {
 	shm_buffer_destroy
 };
 
-static struct wl_shm_buffer *
-wl_shm_buffer_init(struct wl_shm *shm, struct wl_client *client, uint32_t id,
-		   int32_t width, int32_t height,
-		   int32_t stride, uint32_t format, void *data)
-{
-	struct wl_shm_buffer *buffer;
-
-	buffer = calloc(1, sizeof *buffer);
-	if (buffer == NULL)
-		return NULL;
-
-	buffer->buffer.width = width;
-	buffer->buffer.height = height;
-	buffer->format = format;
-	buffer->stride = stride;
-	buffer->data = data;
-
-	buffer->buffer.resource.object.id = id;
-	buffer->buffer.resource.object.interface = &wl_buffer_interface;
-	buffer->buffer.resource.object.implementation = (void (**)(void))
-		&shm_buffer_interface;
-
-	buffer->buffer.resource.data = buffer;
-	buffer->buffer.resource.client = client;
-	buffer->buffer.resource.destroy = destroy_buffer;
-
-	buffer->shm = shm;
-	
-	buffer->shm->callbacks->buffer_created(&buffer->buffer);
-
-	return buffer;
-}
-
 static void
-shm_create_buffer(struct wl_client *client, struct wl_resource *resource,
-		  uint32_t id, int fd, int32_t width, int32_t height,
-		  uint32_t stride, uint32_t format)
+shm_pool_create_buffer(struct wl_client *client, struct wl_resource *resource,
+		       uint32_t id, int32_t offset,
+		       int32_t width, int32_t height,
+		       int32_t stride, uint32_t format)
 {
-	struct wl_shm *shm = resource->data;
+	struct wl_shm_pool *pool = resource->data;
 	struct wl_shm_buffer *buffer;
-	void *data;
-
 
 	switch (format) {
 	case WL_SHM_FORMAT_ARGB8888:
@@ -130,43 +100,138 @@ shm_create_buffer(struct wl_client *client, struct wl_resource *resource,
 		wl_resource_post_error(resource,
 				       WL_SHM_ERROR_INVALID_FORMAT,
 				       "invalid format");
-		close(fd);
 		return;
 	}
 
-	if (width < 0 || height < 0 || stride < width) {
+	if (offset < 0 || width <= 0 || height <= 0 || stride < width ||
+	    INT32_MAX / stride <= height ||
+	    offset > pool->size - stride * height) {
 		wl_resource_post_error(resource,
 				       WL_SHM_ERROR_INVALID_STRIDE,
 				       "invalid width, height or stride (%dx%d, %u)",
 				       width, height, stride);
-		close(fd);
 		return;
 	}
 
-	data = mmap(NULL, stride * height,
-		    PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
-
-	close(fd);
-	if (data == MAP_FAILED) {
-		wl_resource_post_error(resource,
-				       WL_SHM_ERROR_INVALID_FD,
-				       "failed mmap fd %d", fd);
-		return;
-	}
-
-	buffer = wl_shm_buffer_init(shm, client, id,
-				    width, height, stride, format, data);
+	buffer = malloc(sizeof *buffer);
 	if (buffer == NULL) {
-		munmap(data, stride * height);
 		wl_resource_post_no_memory(resource);
 		return;
 	}
 
+	buffer->buffer.width = width;
+	buffer->buffer.height = height;
+	buffer->buffer.busy_count = 0;
+	buffer->format = format;
+	buffer->stride = stride;
+	buffer->offset = offset;
+	buffer->pool = pool;
+	pool->refcount++;
+
+	buffer->buffer.resource.object.id = id;
+	buffer->buffer.resource.object.interface = &wl_buffer_interface;
+	buffer->buffer.resource.object.implementation = (void (**)(void))
+		&shm_buffer_interface;
+
+	buffer->buffer.resource.data = buffer;
+	buffer->buffer.resource.client = resource->client;
+	buffer->buffer.resource.destroy = destroy_buffer;
+
 	wl_client_add_resource(client, &buffer->buffer.resource);
 }
 
-const static struct wl_shm_interface shm_interface = {
-	shm_create_buffer
+static void
+destroy_pool(struct wl_resource *resource)
+{
+	struct wl_shm_pool *pool = resource->data;
+
+	shm_pool_unref(pool);
+}
+
+static void
+shm_pool_destroy(struct wl_client *client, struct wl_resource *resource)
+{
+	wl_resource_destroy(resource);
+}
+
+static void
+shm_pool_resize(struct wl_client *client, struct wl_resource *resource,
+		int32_t size)
+{
+	struct wl_shm_pool *pool = resource->data;
+	void *data;
+
+	data = mremap(pool->data, pool->size, size, MREMAP_MAYMOVE);
+
+	if (data == MAP_FAILED) {
+		wl_resource_post_error(resource,
+				       WL_SHM_ERROR_INVALID_FD,
+				       "failed mremap");
+		return;
+	}
+
+	pool->data = data;
+	pool->size = size;
+}
+
+struct wl_shm_pool_interface shm_pool_interface = {
+	shm_pool_create_buffer,
+	shm_pool_destroy,
+	shm_pool_resize
+};
+
+static void
+shm_create_pool(struct wl_client *client, struct wl_resource *resource,
+		uint32_t id, int fd, int32_t size)
+{
+	struct wl_shm_pool *pool;
+
+	pool = malloc(sizeof *pool);
+	if (pool == NULL) {
+		wl_resource_post_no_memory(resource);
+		goto err_close;
+	}
+
+	if (size <= 0) {
+		wl_resource_post_error(resource,
+				       WL_SHM_ERROR_INVALID_STRIDE,
+				       "invalid size (%d)", size);
+		goto err_free;
+	}
+
+	pool->refcount = 1;
+	pool->size = size;
+	pool->data = mmap(NULL, size,
+			  PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
+	if (pool->data == MAP_FAILED) {
+		wl_resource_post_error(resource,
+				       WL_SHM_ERROR_INVALID_FD,
+				       "failed mmap fd %d", fd);
+		goto err_free;
+	}
+	close(fd);
+
+	pool->resource.object.id = id;
+	pool->resource.object.interface = &wl_shm_pool_interface;
+	pool->resource.object.implementation =
+		(void (**)(void)) &shm_pool_interface;
+
+	pool->resource.data = pool;
+	pool->resource.client = client;
+	pool->resource.destroy = destroy_pool;
+
+	wl_client_add_resource(client, &pool->resource);
+
+	return;
+
+err_close:
+	close(fd);
+err_free:
+	free(pool);
+}
+
+static const struct wl_shm_interface shm_interface = {
+	shm_create_pool
 };
 
 static void
@@ -178,40 +243,58 @@ bind_shm(struct wl_client *client,
 	resource = wl_client_add_object(client, &wl_shm_interface,
 					&shm_interface, id, data);
 
-	wl_resource_post_event(resource, WL_SHM_FORMAT,
-			       WL_SHM_FORMAT_ARGB8888);
-	wl_resource_post_event(resource, WL_SHM_FORMAT,
-			       WL_SHM_FORMAT_XRGB8888);
+	wl_shm_send_format(resource, WL_SHM_FORMAT_ARGB8888);
+	wl_shm_send_format(resource, WL_SHM_FORMAT_XRGB8888);
 }
 
-WL_EXPORT struct wl_shm *
-wl_shm_init(struct wl_display *display,
-	    const struct wl_shm_callbacks *callbacks)
+WL_EXPORT int
+wl_display_init_shm(struct wl_display *display)
 {
-	struct wl_shm *shm;
+	if (!wl_display_add_global(display, &wl_shm_interface, NULL, bind_shm))
+		return -1;
 
-	shm = malloc(sizeof *shm);
-	if (!shm)
-		return NULL;
+	return 0;
+}
 
-	if (!wl_display_add_global(display,
-				   &wl_shm_interface, shm, bind_shm)) {
-
-		free(shm);
+WL_EXPORT struct wl_buffer *
+wl_shm_buffer_create(struct wl_client *client,
+		     uint32_t id, int32_t width, int32_t height,
+		     int32_t stride, uint32_t format)
+{
+	struct wl_shm_buffer *buffer;
+			     
+	switch (format) {
+	case WL_SHM_FORMAT_ARGB8888:
+	case WL_SHM_FORMAT_XRGB8888:
+		break;
+	default:
 		return NULL;
 	}
 
-	shm->callbacks = callbacks;
+	buffer = malloc(sizeof *buffer + stride * height);
+	if (buffer == NULL)
+		return NULL;
 
-	return shm;
-}
+	buffer->buffer.width = width;
+	buffer->buffer.height = height;
+	buffer->buffer.busy_count = 0;
+	buffer->format = format;
+	buffer->stride = stride;
+	buffer->offset = 0;
+	buffer->pool = NULL;
 
-WL_EXPORT void
-wl_shm_finish(struct wl_shm *shm)
-{
-	/* FIXME: add wl_display_del_{object,global} */
+	buffer->buffer.resource.object.id = id;
+	buffer->buffer.resource.object.interface = &wl_buffer_interface;
+	buffer->buffer.resource.object.implementation = (void (**)(void))
+		&shm_buffer_interface;
 
-	free(shm);
+	buffer->buffer.resource.data = buffer;
+	buffer->buffer.resource.client = client;
+	buffer->buffer.resource.destroy = destroy_buffer;
+
+	wl_client_add_resource(client, &buffer->buffer.resource);
+
+	return &buffer->buffer;
 }
 
 WL_EXPORT int
@@ -240,7 +323,10 @@ wl_shm_buffer_get_data(struct wl_buffer *buffer_base)
 	if (!wl_buffer_is_shm(buffer_base))
 		return NULL;
 
-	return buffer->data;
+	if (buffer->pool)
+		return buffer->pool->data + buffer->offset;
+	else
+		return buffer + 1;
 }
 
 WL_EXPORT uint32_t
@@ -250,3 +336,19 @@ wl_shm_buffer_get_format(struct wl_buffer *buffer_base)
 
 	return buffer->format;
 }
+
+WL_EXPORT int32_t
+wl_shm_buffer_get_width(struct wl_buffer *buffer_base)
+{
+	struct wl_shm_buffer *buffer = (struct wl_shm_buffer *) buffer_base;
+
+	return buffer->buffer.width;
+}
+
+WL_EXPORT int32_t
+wl_shm_buffer_get_height(struct wl_buffer *buffer_base)
+{
+	struct wl_shm_buffer *buffer = (struct wl_shm_buffer *) buffer_base;
+
+	return buffer->buffer.height;
+}
